@@ -21,6 +21,7 @@ import time
 import uuid
 
 import cachetools
+from oslo_utils import timeutils
 from six import moves
 
 import oslo_messaging
@@ -34,11 +35,27 @@ from oslo_messaging._i18n import _LW
 
 LOG = logging.getLogger(__name__)
 
+# Minimum/Maximum sleep between a poll and ack/requeue
+# Maximum should be small enough to not get rejected ack,
+# minimum should be big enough to not burn the CPU.
+ACK_REQUEUE_EVERY_SECONDS_MIN = 0.001
+ACK_REQUEUE_EVERY_SECONDS_MAX = 1.0
+
+
+def do_pending_tasks(tasks):
+    while True:
+        try:
+            task = tasks.get(block=False)
+        except moves.queue.Empty:
+            break
+        else:
+            task()
+
 
 class AMQPIncomingMessage(base.RpcIncomingMessage):
 
     def __init__(self, listener, ctxt, message, unique_id, msg_id, reply_q,
-                 obsolete_reply_queues):
+                 obsolete_reply_queues, pending_message_actions):
         super(AMQPIncomingMessage, self).__init__(ctxt, message)
         self.listener = listener
 
@@ -46,15 +63,17 @@ class AMQPIncomingMessage(base.RpcIncomingMessage):
         self.msg_id = msg_id
         self.reply_q = reply_q
         self._obsolete_reply_queues = obsolete_reply_queues
+        self._pending_tasks = pending_message_actions
+        self.stopwatch = timeutils.StopWatch()
+        self.stopwatch.start()
 
-    def _send_reply(self, conn, reply=None, failure=None, log_failure=True):
+    def _send_reply(self, conn, reply=None, failure=None):
         if not self._obsolete_reply_queues.reply_q_valid(self.reply_q,
                                                          self.msg_id):
             return
 
         if failure:
-            failure = rpc_common.serialize_remote_exception(failure,
-                                                            log_failure)
+            failure = rpc_common.serialize_remote_exception(failure)
         # NOTE(sileht): ending can be removed in N*, see Listener.wait()
         # for more detail.
         msg = {'result': reply, 'failure': failure, 'ending': True,
@@ -63,13 +82,15 @@ class AMQPIncomingMessage(base.RpcIncomingMessage):
         unique_id = msg[rpc_amqp.UNIQUE_ID]
 
         LOG.debug("sending reply msg_id: %(msg_id)s "
-                  "reply queue: %(reply_q)s", {
+                  "reply queue: %(reply_q)s "
+                  "time elapsed: %(elapsed)ss", {
                       'msg_id': self.msg_id,
                       'unique_id': unique_id,
-                      'reply_q': self.reply_q})
+                      'reply_q': self.reply_q,
+                      'elapsed': self.stopwatch.elapsed()})
         conn.direct_send(self.reply_q, rpc_common.serialize_msg(msg))
 
-    def reply(self, reply=None, failure=None, log_failure=True):
+    def reply(self, reply=None, failure=None):
         if not self.msg_id:
             # NOTE(Alexei_987) not sending reply, if msg_id is empty
             #    because reply should not be expected by caller side
@@ -91,8 +112,7 @@ class AMQPIncomingMessage(base.RpcIncomingMessage):
             try:
                 with self.listener.driver._get_connection(
                         rpc_common.PURPOSE_SEND) as conn:
-                    self._send_reply(conn, reply, failure,
-                                     log_failure=log_failure)
+                    self._send_reply(conn, reply, failure)
                 return
             except rpc_amqp.AMQPDestinationNotFound:
                 if timer.check_return() > 0:
@@ -113,7 +133,7 @@ class AMQPIncomingMessage(base.RpcIncomingMessage):
                     return
 
     def acknowledge(self):
-        self.message.acknowledge()
+        self._pending_tasks.put(self.message.acknowledge)
         self.listener.msg_id_cache.add(self.unique_id)
 
     def requeue(self):
@@ -123,7 +143,7 @@ class AMQPIncomingMessage(base.RpcIncomingMessage):
         # msg_id_cache, the message will be reconsumed, the only difference is
         # the message stay at the beginning of the queue instead of moving to
         # the end.
-        self.message.requeue()
+        self._pending_tasks.put(self.message.requeue)
 
 
 class ObsoleteReplyQueuesCache(object):
@@ -171,7 +191,7 @@ class ObsoleteReplyQueuesCache(object):
                                         'msg_id': msg_id})
 
 
-class AMQPListener(base.Listener):
+class AMQPListener(base.PollStyleListener):
 
     def __init__(self, driver, conn):
         super(AMQPListener, self).__init__(driver.prefetch_size)
@@ -181,41 +201,57 @@ class AMQPListener(base.Listener):
         self.incoming = []
         self._stopped = threading.Event()
         self._obsolete_reply_queues = ObsoleteReplyQueuesCache()
+        self._pending_tasks = moves.queue.Queue()
+        self._current_timeout = ACK_REQUEUE_EVERY_SECONDS_MIN
 
     def __call__(self, message):
         ctxt = rpc_amqp.unpack_context(message)
-
-        # FIXME(sileht): Don't log the message until strutils is more
-        # efficient, (rpc_amqp.unpack_context already log the context)
-        # LOG.debug(u'received: %s',
-        #           strutils.mask_password(six.text_type(dict(message))))
-
         unique_id = self.msg_id_cache.check_duplicate_message(message)
+        if ctxt.msg_id:
+            LOG.debug("received message msg_id: %(msg_id)s reply to "
+                      "%(queue)s", {'queue': ctxt.reply_q,
+                                    'msg_id': ctxt.msg_id})
+        else:
+            LOG.debug("received message with unique_id: %s", unique_id)
 
-        LOG.debug("received message msg_id: %(msg_id)s reply to %(queue)s", {
-            'queue': ctxt.reply_q, 'msg_id': ctxt.msg_id})
-
-        self.incoming.append(AMQPIncomingMessage(self,
-                                                 ctxt.to_dict(),
-                                                 message,
-                                                 unique_id,
-                                                 ctxt.msg_id,
-                                                 ctxt.reply_q,
-                                                 self._obsolete_reply_queues))
+        self.incoming.append(AMQPIncomingMessage(
+            self,
+            ctxt.to_dict(),
+            message,
+            unique_id,
+            ctxt.msg_id,
+            ctxt.reply_q,
+            self._obsolete_reply_queues,
+            self._pending_tasks))
 
     @base.batch_poll_helper
     def poll(self, timeout=None):
+        stopwatch = timeutils.StopWatch(duration=timeout).start()
+
         while not self._stopped.is_set():
+            do_pending_tasks(self._pending_tasks)
+
             if self.incoming:
                 return self.incoming.pop(0)
-            try:
-                self.conn.consume(timeout=timeout)
-            except rpc_common.Timeout:
+
+            left = stopwatch.leftover(return_none=True)
+            if left is None:
+                left = self._current_timeout
+            if left <= 0:
                 return None
 
+            try:
+                self.conn.consume(timeout=min(self._current_timeout, left))
+            except rpc_common.Timeout:
+                self._current_timeout = max(self._current_timeout * 2,
+                                            ACK_REQUEUE_EVERY_SECONDS_MAX)
+            else:
+                self._current_timeout = ACK_REQUEUE_EVERY_SECONDS_MIN
+
     def stop(self):
         self._stopped.set()
         self.conn.stop_consuming()
+        do_pending_tasks(self._pending_tasks)
 
     def cleanup(self):
         # Closes listener connection
@@ -249,11 +285,14 @@ class ReplyWaiters(object):
 
     def add(self, msg_id):
         self._queues[msg_id] = moves.queue.Queue()
-        if len(self._queues) > self._wrn_threshold:
-            LOG.warning(_LW('Number of call queues is greater than warning '
-                            'threshold: %(old_threshold)s. There could be a '
-                            'leak. Increasing threshold to: %(threshold)s'),
-                        {'old_threshold': self._wrn_threshold,
+        queues_length = len(self._queues)
+        if queues_length > self._wrn_threshold:
+            LOG.warning(_LW('Number of call queues is %(queues_length)s, '
+                            'greater than warning threshold: %(old_threshold)s'
+                            '. There could be a leak. Increasing threshold to:'
+                            ' %(threshold)s'),
+                        {'queues_length': queues_length,
+                         'old_threshold': self._wrn_threshold,
                          'threshold': self._wrn_threshold * 2})
             self._wrn_threshold *= 2
 
@@ -267,6 +306,7 @@ class ReplyWaiter(object):
         self.allowed_remote_exmods = allowed_remote_exmods
         self.msg_id_cache = rpc_amqp._MsgIdCache()
         self.waiters = ReplyWaiters()
+        self._pending_tasks = moves.queue.Queue()
 
         self.conn.declare_direct_consumer(reply_q, self)
 
@@ -281,17 +321,26 @@ class ReplyWaiter(object):
             self.conn.stop_consuming()
             self._thread.join()
             self._thread = None
+            do_pending_tasks(self._pending_tasks)
 
     def poll(self):
+        current_timeout = ACK_REQUEUE_EVERY_SECONDS_MIN
         while not self._thread_exit_event.is_set():
+            do_pending_tasks(self._pending_tasks)
             try:
-                self.conn.consume()
+                # ack every ACK_REQUEUE_EVERY_SECONDS_MAX seconds
+                self.conn.consume(timeout=current_timeout)
+            except rpc_common.Timeout:
+                current_timeout = max(current_timeout * 2,
+                                      ACK_REQUEUE_EVERY_SECONDS_MAX)
             except Exception:
                 LOG.exception(_LE("Failed to process incoming message, "
                               "retrying..."))
+            else:
+                current_timeout = ACK_REQUEUE_EVERY_SECONDS_MIN
 
     def __call__(self, message):
-        message.acknowledge()
+        self._pending_tasks.put(message.acknowledge)
         incoming_msg_id = message.pop('_msg_id', None)
         if message.get('ending'):
             LOG.debug("received reply msg_id: %s", incoming_msg_id)
@@ -392,15 +441,6 @@ class AMQPDriverBase(base.BaseDriver):
               wait_for_reply=None, timeout=None,
               envelope=True, notify=False, retry=None):
 
-        # FIXME(markmc): remove this temporary hack
-        class Context(object):
-            def __init__(self, d):
-                self.d = d
-
-            def to_dict(self):
-                return self.d
-
-        context = Context(ctxt)
         msg = message
 
         if wait_for_reply:
@@ -411,7 +451,7 @@ class AMQPDriverBase(base.BaseDriver):
         rpc_amqp._add_unique_id(msg)
         unique_id = msg[rpc_amqp.UNIQUE_ID]
 
-        rpc_amqp.pack_context(msg, context)
+        rpc_amqp.pack_context(msg, ctxt)
 
         if envelope:
             msg = rpc_common.serialize_msg(msg)
@@ -445,7 +485,7 @@ class AMQPDriverBase(base.BaseDriver):
                     log_msg += "exchange '%(exchange)s'" \
                                " topic '%(topic)s'" % {
                                    'exchange': exchange,
-                                   'topic': target.topic}
+                                   'topic': topic}
                     LOG.debug(log_msg)
                     conn.topic_send(exchange_name=exchange, topic=topic,
                                     msg=msg, timeout=timeout, retry=retry)
@@ -468,7 +508,7 @@ class AMQPDriverBase(base.BaseDriver):
         return self._send(target, ctxt, message,
                           envelope=(version == 2.0), notify=True, retry=retry)
 
-    def listen(self, target):
+    def listen(self, target, batch_size, batch_timeout):
         conn = self._get_connection(rpc_common.PURPOSE_LISTEN)
 
         listener = AMQPListener(self, conn)
@@ -482,10 +522,18 @@ class AMQPDriverBase(base.BaseDriver):
                                     callback=listener)
         conn.declare_fanout_consumer(target.topic, listener)
 
-        return listener
+        return base.PollStyleListenerAdapter(listener, batch_size,
+                                             batch_timeout)
 
-    def listen_for_notifications(self, targets_and_priorities, pool):
+    def listen_for_notifications(self, targets_and_priorities, pool,
+                                 batch_size, batch_timeout):
         conn = self._get_connection(rpc_common.PURPOSE_LISTEN)
+        # NOTE(sileht): The application set batch_size, so we don't need to
+        # prefetch more messages, especially for notifications. Notifications
+        # queues can be really big when the consumer have disapear during a
+        # long period, and when it come back, kombu/pyamqp will fetch all
+        # messages it can. So we override the default qos prefetch value
+        conn.connection.rabbit_qos_prefetch_count = batch_size
 
         listener = AMQPListener(self, conn)
         for target, priority in targets_and_priorities:
@@ -493,7 +541,8 @@ class AMQPDriverBase(base.BaseDriver):
                 exchange_name=self._get_exchange(target),
                 topic='%s.%s' % (target.topic, priority),
                 callback=listener, queue_name=pool)
-        return listener
+        return base.PollStyleListenerAdapter(listener, batch_size,
+                                             batch_timeout)
 
     def cleanup(self):
         if self._connection_pool:
