@@ -12,13 +12,16 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-import collections
 import contextlib
+import errno
 import functools
+import itertools
+import math
 import os
 import random
 import socket
 import ssl
+import sys
 import threading
 import time
 import uuid
@@ -29,6 +32,7 @@ import kombu.entity
 import kombu.messaging
 from oslo_config import cfg
 from oslo_log import log as logging
+from oslo_utils import eventletutils
 from oslo_utils import netutils
 import six
 from six.moves.urllib import parse
@@ -45,27 +49,34 @@ from oslo_messaging._i18n import _LW
 from oslo_messaging import _utils
 from oslo_messaging import exceptions
 
+# NOTE(sileht): don't exists in py2 socket module
+TCP_USER_TIMEOUT = 18
+
 
 rabbit_opts = [
-    cfg.StrOpt('kombu_ssl_version',
+    cfg.BoolOpt('ssl',
+                default=False,
+                deprecated_name='rabbit_use_ssl',
+                help='Connect over SSL.'),
+    cfg.StrOpt('ssl_version',
                default='',
-               deprecated_group='DEFAULT',
+               deprecated_name='kombu_ssl_version',
                help='SSL version to use (valid only if SSL enabled). '
                     'Valid values are TLSv1 and SSLv23. SSLv2, SSLv3, '
                     'TLSv1_1, and TLSv1_2 may be available on some '
                     'distributions.'
                ),
-    cfg.StrOpt('kombu_ssl_keyfile',
+    cfg.StrOpt('ssl_key_file',
                default='',
-               deprecated_group='DEFAULT',
+               deprecated_name='kombu_ssl_keyfile',
                help='SSL key file (valid only if SSL enabled).'),
-    cfg.StrOpt('kombu_ssl_certfile',
+    cfg.StrOpt('ssl_cert_file',
                default='',
-               deprecated_group='DEFAULT',
+               deprecated_name='kombu_ssl_certfile',
                help='SSL cert file (valid only if SSL enabled).'),
-    cfg.StrOpt('kombu_ssl_ca_certs',
+    cfg.StrOpt('ssl_ca_file',
                default='',
-               deprecated_group='DEFAULT',
+               deprecated_name='kombu_ssl_ca_certs',
                help='SSL certification authority file '
                     '(valid only if SSL enabled).'),
     cfg.FloatOpt('kombu_reconnect_delay',
@@ -75,12 +86,12 @@ rabbit_opts = [
                       'AMQP consumer cancel notification.'),
     cfg.StrOpt('kombu_compression',
                help="EXPERIMENTAL: Possible values are: gzip, bz2. If not "
-                    "set compression will not be used. This option may not"
+                    "set compression will not be used. This option may not "
                     "be available in future versions."),
     cfg.IntOpt('kombu_missing_consumer_retry_timeout',
                deprecated_name="kombu_reconnect_timeout",
                default=60,
-               help='How long to wait a missing client beforce abandoning to '
+               help='How long to wait a missing client before abandoning to '
                     'send it its replies. This value should not be longer '
                     'than rpc_response_timeout.'),
     cfg.StrOpt('kombu_failover_strategy',
@@ -93,36 +104,45 @@ rabbit_opts = [
     cfg.StrOpt('rabbit_host',
                default='localhost',
                deprecated_group='DEFAULT',
+               deprecated_for_removal=True,
+               deprecated_reason="Replaced by [DEFAULT]/transport_url",
                help='The RabbitMQ broker address where a single node is '
                     'used.'),
     cfg.PortOpt('rabbit_port',
                 default=5672,
                 deprecated_group='DEFAULT',
+                deprecated_for_removal=True,
+                deprecated_reason="Replaced by [DEFAULT]/transport_url",
                 help='The RabbitMQ broker port where a single node is used.'),
     cfg.ListOpt('rabbit_hosts',
                 default=['$rabbit_host:$rabbit_port'],
                 deprecated_group='DEFAULT',
+                deprecated_for_removal=True,
+                deprecated_reason="Replaced by [DEFAULT]/transport_url",
                 help='RabbitMQ HA cluster host:port pairs.'),
-    cfg.BoolOpt('rabbit_use_ssl',
-                default=False,
-                deprecated_group='DEFAULT',
-                help='Connect over SSL for RabbitMQ.'),
     cfg.StrOpt('rabbit_userid',
                default='guest',
                deprecated_group='DEFAULT',
+               deprecated_for_removal=True,
+               deprecated_reason="Replaced by [DEFAULT]/transport_url",
                help='The RabbitMQ userid.'),
     cfg.StrOpt('rabbit_password',
                default='guest',
                deprecated_group='DEFAULT',
+               deprecated_for_removal=True,
+               deprecated_reason="Replaced by [DEFAULT]/transport_url",
                help='The RabbitMQ password.',
                secret=True),
     cfg.StrOpt('rabbit_login_method',
+               choices=('PLAIN', 'AMQPLAIN', 'RABBIT-CR-DEMO'),
                default='AMQPLAIN',
                deprecated_group='DEFAULT',
                help='The RabbitMQ login method.'),
     cfg.StrOpt('rabbit_virtual_host',
                default='/',
                deprecated_group='DEFAULT',
+               deprecated_for_removal=True,
+               deprecated_reason="Replaced by [DEFAULT]/transport_url",
                help='The RabbitMQ virtual host.'),
     cfg.IntOpt('rabbit_retry_interval',
                default=1,
@@ -138,18 +158,25 @@ rabbit_opts = [
                     'Default is 30 seconds.'),
     cfg.IntOpt('rabbit_max_retries',
                default=0,
+               deprecated_for_removal=True,
                deprecated_group='DEFAULT',
                help='Maximum number of RabbitMQ connection retries. '
                     'Default is 0 (infinite retry count).'),
     cfg.BoolOpt('rabbit_ha_queues',
                 default=False,
                 deprecated_group='DEFAULT',
-                help='Use HA queues in RabbitMQ (x-ha-policy: all). '
-                     'If you change this option, you must wipe the '
-                     'RabbitMQ database.'),
+                help='Try to use HA queues in RabbitMQ (x-ha-policy: all). '
+                'If you change this option, you must wipe the RabbitMQ '
+                'database. In RabbitMQ 3.0, queue mirroring is no longer '
+                'controlled by the x-ha-policy argument when declaring a '
+                'queue. If you just want to make sure that all queues (except '
+                'those with auto-generated names) are mirrored across all '
+                'nodes, run: '
+                """\"rabbitmqctl set_policy HA '^(?!amq\.).*' """
+                """'{"ha-mode": "all"}' \""""),
     cfg.IntOpt('rabbit_transient_queues_ttl',
                min=1,
-               default=600,
+               default=1800,
                help='Positive integer representing duration in seconds for '
                     'queue TTL (x-expires). Queues which are unused for the '
                     'duration of the TTL are automatically deleted. The '
@@ -182,13 +209,17 @@ LOG = logging.getLogger(__name__)
 def _get_queue_arguments(rabbit_ha_queues, rabbit_queue_ttl):
     """Construct the arguments for declaring a queue.
 
-    If the rabbit_ha_queues option is set, we declare a mirrored queue
+    If the rabbit_ha_queues option is set, we try to declare a mirrored queue
     as described here:
 
       http://www.rabbitmq.com/ha.html
 
     Setting x-ha-policy to all means that the queue will be mirrored
-    to all nodes in the cluster.
+    to all nodes in the cluster. In RabbitMQ 3.0, queue mirroring is
+    no longer controlled by the x-ha-policy argument when declaring a
+    queue. If you just want to make sure that all queues (except those
+    with auto-generated names) are mirrored across all nodes, run:
+      rabbitmqctl set_policy HA '^(?!amq\.).*' '{"ha-mode": "all"}'
 
     If the rabbit_queue_ttl option is > 0, then the queue is
     declared with the "Queue TTL" value as described here:
@@ -231,8 +262,8 @@ class Consumer(object):
 
     def __init__(self, exchange_name, queue_name, routing_key, type, durable,
                  exchange_auto_delete, queue_auto_delete, callback,
-                 nowait=True, rabbit_ha_queues=None, rabbit_queue_ttl=0):
-        """Init the Publisher class with the exchange_name, routing_key,
+                 nowait=False, rabbit_ha_queues=None, rabbit_queue_ttl=0):
+        """Init the Consumer class with the exchange_name, routing_key,
         type, durable auto_delete
         """
         self.queue_name = queue_name
@@ -248,6 +279,7 @@ class Consumer(object):
                                                     rabbit_queue_ttl)
 
         self.queue = None
+        self._declared_on = None
         self.exchange = kombu.entity.Exchange(
             name=exchange_name,
             type=type,
@@ -256,6 +288,7 @@ class Consumer(object):
 
     def declare(self, conn):
         """Re-declare the queue after a rabbit (re)connect."""
+
         self.queue = kombu.entity.Queue(
             name=self.queue_name,
             channel=conn.channel,
@@ -266,8 +299,8 @@ class Consumer(object):
             queue_arguments=self.queue_arguments)
 
         try:
-            LOG.trace('ConsumerBase.declare: '
-                      'queue %s', self.queue_name)
+            LOG.debug('[%s] Queue.declare: %s',
+                      conn.connection_id, self.queue_name)
             self.queue.declare()
         except conn.connection.channel_errors as exc:
             # NOTE(jrosenboom): This exception may be triggered by a race
@@ -279,17 +312,44 @@ class Consumer(object):
                 self.queue.declare()
             else:
                 raise
+        self._declared_on = conn.channel
 
-    def consume(self, tag):
+    def consume(self, conn, tag):
         """Actually declare the consumer on the amqp channel.  This will
         start the flow of messages from the queue.  Using the
         Connection.consume() will process the messages,
         calling the appropriate callback.
         """
 
-        self.queue.consume(callback=self._callback,
-                           consumer_tag=six.text_type(tag),
-                           nowait=self.nowait)
+        # Ensure we are on the correct channel before consuming
+        if conn.channel != self._declared_on:
+            self.declare(conn)
+        try:
+            self.queue.consume(callback=self._callback,
+                               consumer_tag=six.text_type(tag),
+                               nowait=self.nowait)
+        except conn.connection.channel_errors as exc:
+            # We retries once because of some races that we can
+            # recover before informing the deployer
+            # bugs.launchpad.net/oslo.messaging/+bug/1581148
+            # bugs.launchpad.net/oslo.messaging/+bug/1609766
+            # bugs.launchpad.net/neutron/+bug/1318721
+
+            # 406 error code relates to messages that are doubled ack'd
+
+            # At any channel error, the RabbitMQ closes
+            # the channel, but the amqp-lib quietly re-open
+            # it. So, we must reset all tags and declare
+            # all consumers again.
+            conn._new_tags = set(conn._consumers.values())
+            if exc.code == 404 or (exc.code == 406 and
+                                   exc.method_name == 'Basic.ack'):
+                self.declare(conn)
+                self.queue.consume(callback=self._callback,
+                                   consumer_tag=six.text_type(tag),
+                                   nowait=self.nowait)
+            else:
+                raise
 
     def cancel(self, tag):
         LOG.trace('ConsumerBase.cancel: canceling %s', tag)
@@ -309,7 +369,7 @@ class Consumer(object):
         except Exception:
             LOG.exception(_LE("Failed to process message"
                               " ... skipping it."))
-            message.ack()
+            message.reject()
 
 
 class DummyConnectionLock(_utils.DummyLock):
@@ -318,9 +378,9 @@ class DummyConnectionLock(_utils.DummyLock):
 
 
 class ConnectionLock(DummyConnectionLock):
-    """Lock object to protect access the the kombu connection
+    """Lock object to protect access to the kombu connection
 
-    This is a lock object to protect access the the kombu connection
+    This is a lock object to protect access to the kombu connection
     object between the heartbeat thread and the driver thread.
 
     They are two way to acquire this lock:
@@ -334,7 +394,7 @@ class ConnectionLock(DummyConnectionLock):
     starvation when the driver sends a lot of messages.
 
     So when lock.heartbeat_acquire() is called next time the lock
-    is released(), the caller unconditionnaly acquires
+    is released(), the caller unconditionally acquires
     the lock, even someone else have asked for the lock before it.
     """
 
@@ -345,7 +405,7 @@ class ConnectionLock(DummyConnectionLock):
         self._monitor = threading.Lock()
         self._workers_locks = threading.Condition(self._monitor)
         self._heartbeat_lock = threading.Condition(self._monitor)
-        self._get_thread_id = _utils.fetch_current_thread_functor()
+        self._get_thread_id = eventletutils.fetch_current_thread_functor()
 
     def acquire(self):
         with self._monitor:
@@ -419,17 +479,17 @@ class Connection(object):
         self.kombu_reconnect_delay = driver_conf.kombu_reconnect_delay
         self.amqp_durable_queues = driver_conf.amqp_durable_queues
         self.amqp_auto_delete = driver_conf.amqp_auto_delete
-        self.rabbit_use_ssl = driver_conf.rabbit_use_ssl
+        self.ssl = driver_conf.ssl
         self.kombu_missing_consumer_retry_timeout = \
             driver_conf.kombu_missing_consumer_retry_timeout
         self.kombu_failover_strategy = driver_conf.kombu_failover_strategy
         self.kombu_compression = driver_conf.kombu_compression
 
-        if self.rabbit_use_ssl:
-            self.kombu_ssl_version = driver_conf.kombu_ssl_version
-            self.kombu_ssl_keyfile = driver_conf.kombu_ssl_keyfile
-            self.kombu_ssl_certfile = driver_conf.kombu_ssl_certfile
-            self.kombu_ssl_ca_certs = driver_conf.kombu_ssl_ca_certs
+        if self.ssl:
+            self.ssl_version = driver_conf.ssl_version
+            self.ssl_key_file = driver_conf.ssl_key_file
+            self.ssl_cert_file = driver_conf.ssl_cert_file
+            self.ssl_ca_file = driver_conf.ssl_ca_file
 
         # Try forever?
         if self.max_retries <= 0:
@@ -485,8 +545,18 @@ class Connection(object):
 
         self._initial_pid = os.getpid()
 
-        self._consumers = []
-        self._new_consumers = []
+        self._consumers = {}
+        self._producer = None
+        self._new_tags = set()
+        self._active_tags = {}
+        self._tags = itertools.count(1)
+
+        # Set of exchanges and queues declared on the channel to avoid
+        # unnecessary redeclaration. This set is resetted each time
+        # the connection is resetted in Connection._set_current_channel
+        self._declared_exchanges = set()
+        self._declared_queues = set()
+
         self._consume_loop_stopped = False
         self.channel = None
         self.purpose = purpose
@@ -499,6 +569,10 @@ class Connection(object):
         else:
             self._connection_lock = DummyConnectionLock()
 
+        self.connection_id = str(uuid.uuid4())
+        self.name = "%s:%d:%s" % (os.path.basename(sys.argv[0]),
+                                  os.getpid(),
+                                  self.connection_id)
         self.connection = kombu.connection.Connection(
             self._url, ssl=self._fetch_ssl_params(),
             login_method=self.login_method,
@@ -506,13 +580,21 @@ class Connection(object):
             failover_strategy=self.kombu_failover_strategy,
             transport_options={
                 'confirm_publish': True,
+                'client_properties': {
+                    'capabilities': {
+                        'authentication_failure_close': True,
+                        'connection.blocked': True,
+                        'consumer_cancel_notify': True
+                    },
+                    'connection_name': self.name},
                 'on_blocked': self._on_connection_blocked,
                 'on_unblocked': self._on_connection_unblocked,
             },
         )
 
-        LOG.debug('Connecting to AMQP server on %(hostname)s:%(port)s',
-                  self.connection.info())
+        LOG.debug('[%(connection_id)s] Connecting to AMQP server on'
+                  ' %(hostname)s:%(port)s',
+                  self._get_connection_info())
 
         # NOTE(sileht): kombu recommend to run heartbeat_check every
         # seconds, but we use a lock around the kombu connection
@@ -528,7 +610,8 @@ class Connection(object):
         self._heartbeat_support_log_emitted = False
 
         # NOTE(sileht): just ensure the connection is setuped at startup
-        self.ensure_connection()
+        with self._connection_lock:
+            self.ensure_connection()
 
         # NOTE(sileht): if purpose is PURPOSE_LISTEN
         # the consume code does the heartbeat stuff
@@ -537,9 +620,10 @@ class Connection(object):
         if purpose == rpc_common.PURPOSE_SEND:
             self._heartbeat_start()
 
-        LOG.debug('Connected to AMQP server on %(hostname)s:%(port)s '
-                  'via [%(transport)s] client',
-                  self.connection.info())
+        LOG.debug('[%(connection_id)s] Connected to AMQP server on '
+                  '%(hostname)s:%(port)s via [%(transport)s] client with'
+                  ' port %(client_port)s.',
+                  self._get_connection_info())
 
         # NOTE(sileht): value chosen according the best practice from kombu
         # http://kombu.readthedocs.org/en/latest/reference/kombu.common.html#kombu.common.eventloop
@@ -595,19 +679,19 @@ class Connection(object):
         """Handles fetching what ssl params should be used for the connection
         (if any).
         """
-        if self.rabbit_use_ssl:
+        if self.ssl:
             ssl_params = dict()
 
             # http://docs.python.org/library/ssl.html - ssl.wrap_socket
-            if self.kombu_ssl_version:
+            if self.ssl_version:
                 ssl_params['ssl_version'] = self.validate_ssl_version(
-                    self.kombu_ssl_version)
-            if self.kombu_ssl_keyfile:
-                ssl_params['keyfile'] = self.kombu_ssl_keyfile
-            if self.kombu_ssl_certfile:
-                ssl_params['certfile'] = self.kombu_ssl_certfile
-            if self.kombu_ssl_ca_certs:
-                ssl_params['ca_certs'] = self.kombu_ssl_ca_certs
+                    self.ssl_version)
+            if self.ssl_key_file:
+                ssl_params['keyfile'] = self.ssl_key_file
+            if self.ssl_cert_file:
+                ssl_params['certfile'] = self.ssl_cert_file
+            if self.ssl_ca_file:
+                ssl_params['ca_certs'] = self.ssl_ca_file
                 # We might want to allow variations in the
                 # future with this?
                 ssl_params['cert_reqs'] = ssl.CERT_REQUIRED
@@ -626,7 +710,8 @@ class Connection(object):
         # NOTE(sileht): we reset the channel and ensure
         # the kombu underlying connection works
         self._set_current_channel(None)
-        self.ensure(method=lambda: self.connection.connection)
+        self.ensure(method=self.connection.connect)
+        self.set_transport_socket_timeout()
 
     def ensure(self, method, retry=None,
                recoverable_error_callback=None, error_callback=None,
@@ -654,7 +739,8 @@ class Connection(object):
             retry = None
 
         def on_error(exc, interval):
-            LOG.debug("Received recoverable error from kombu:",
+            LOG.debug("[%s] Received recoverable error from kombu:"
+                      % self.connection_id,
                       exc_info=True)
 
             recoverable_error_callback and recoverable_error_callback(exc)
@@ -664,16 +750,19 @@ class Connection(object):
                         else interval)
 
             info = {'err_str': exc, 'sleep_time': interval}
-            info.update(self.connection.info())
+            info.update(self._get_connection_info())
 
             if 'Socket closed' in six.text_type(exc):
-                LOG.error(_LE('AMQP server %(hostname)s:%(port)s closed'
+                LOG.error(_LE('[%(connection_id)s] AMQP server'
+                              ' %(hostname)s:%(port)s closed'
                               ' the connection. Check login credentials:'
                               ' %(err_str)s'), info)
             else:
-                LOG.error(_LE('AMQP server on %(hostname)s:%(port)s is '
-                              'unreachable: %(err_str)s. Trying again in '
-                              '%(sleep_time)d seconds.'), info)
+                LOG.error(_LE('[%(connection_id)s] AMQP server on '
+                              '%(hostname)s:%(port)s is unreachable: '
+                              '%(err_str)s. Trying again in '
+                              '%(sleep_time)d seconds. Client port: '
+                              '%(client_port)s'), info)
 
             # XXX(nic): when reconnecting to a RabbitMQ cluster
             # with mirrored queues in use, the attempt to release the
@@ -695,30 +784,17 @@ class Connection(object):
             a new channel, we use it the reconfigure our consumers.
             """
             self._set_current_channel(new_channel)
-            for consumer in self._consumers:
-                consumer.declare(self)
+            self.set_transport_socket_timeout()
 
-            LOG.info(_LI('Reconnected to AMQP server on '
-                         '%(hostname)s:%(port)s via [%(transport)s] client'),
-                     self.connection.info())
+            LOG.info(_LI('[%(connection_id)s] Reconnected to AMQP server on '
+                         '%(hostname)s:%(port)s via [%(transport)s] client '
+                         'with port %(client_port)s.'),
+                     self._get_connection_info())
 
         def execute_method(channel):
             self._set_current_channel(channel)
             method()
 
-        # NOTE(sileht): Some dummy driver like the in-memory one doesn't
-        # have notion of recoverable connection, so we must raise the original
-        # exception like kombu does in this case.
-        has_modern_errors = hasattr(
-            self.connection.transport, 'recoverable_connection_errors',
-        )
-        if has_modern_errors:
-            recoverable_errors = (
-                self.connection.recoverable_channel_errors +
-                self.connection.recoverable_connection_errors)
-        else:
-            recoverable_errors = ()
-
         try:
             autoretry_method = self.connection.autoretry(
                 execute_method, channel=self.channel,
@@ -731,7 +807,7 @@ class Connection(object):
             ret, channel = autoretry_method()
             self._set_current_channel(channel)
             return ret
-        except recoverable_errors as exc:
+        except kombu.exceptions.OperationalError as exc:
             LOG.debug("Received recoverable error from kombu:",
                       exc_info=True)
             error_callback and error_callback(exc)
@@ -762,14 +838,18 @@ class Connection(object):
             return
 
         if self.channel is not None:
-            self.PUBLISHER_DECLARED_QUEUES.pop(self.channel, None)
+            self._declared_queues.clear()
+            self._declared_exchanges.clear()
             self.connection.maybe_close_channel(self.channel)
 
         self.channel = new_channel
 
-        if (new_channel is not None and
-           self.purpose == rpc_common.PURPOSE_LISTEN):
-            self._set_qos(new_channel)
+        if new_channel is not None:
+            if self.purpose == rpc_common.PURPOSE_LISTEN:
+                self._set_qos(new_channel)
+            self._producer = kombu.messaging.Producer(new_channel)
+            for consumer in self._consumers:
+                consumer.declare(self)
 
     def _set_qos(self, channel):
         """Set QoS prefetch count on the channel"""
@@ -782,22 +862,27 @@ class Connection(object):
         """Close/release this connection."""
         self._heartbeat_stop()
         if self.connection:
+            for consumer, tag in self._consumers.items():
+                if consumer.type == 'fanout':
+                    LOG.debug('[connection close] Deleting fanout '
+                              'queue: %s ' % consumer.queue.name)
+                    consumer.queue.delete()
             self._set_current_channel(None)
             self.connection.release()
             self.connection = None
 
     def reset(self):
         """Reset a connection so it can be used again."""
-        recoverable_errors = (self.connection.recoverable_channel_errors +
-                              self.connection.recoverable_connection_errors)
-
         with self._connection_lock:
             try:
-                for tag, consumer in enumerate(self._consumers):
+                for consumer, tag in self._consumers.items():
                     consumer.cancel(tag=tag)
-            except recoverable_errors:
+            except kombu.exceptions.OperationalError:
                 self.ensure_connection()
-            self._consumers = []
+            self._consumers.clear()
+            self._active_tags.clear()
+            self._new_tags.clear()
+            self._tags = itertools.count(1)
 
     def _heartbeat_supported_and_enabled(self):
         if self.heartbeat_timeout_threshold <= 0:
@@ -811,8 +896,7 @@ class Connection(object):
             self._heartbeat_support_log_emitted = True
         return False
 
-    @contextlib.contextmanager
-    def _transport_socket_timeout(self, timeout):
+    def set_transport_socket_timeout(self, timeout=None):
         # NOTE(sileht): they are some case where the heartbeat check
         # or the producer.send return only when the system socket
         # timeout if reach. kombu doesn't allow use to customise this
@@ -821,31 +905,57 @@ class Connection(object):
         # kombu==3.0.33. Once the commit below is released, we should
         # try to set the socket timeout in the constructor:
         # https://github.com/celery/py-amqp/pull/64
+
+        heartbeat_timeout = self.heartbeat_timeout_threshold
+        if self._heartbeat_supported_and_enabled():
+            # NOTE(sileht): we are supposed to send heartbeat every
+            # heartbeat_timeout, no need to wait more otherwise will
+            # disconnect us, so raise timeout earlier ourself
+            if timeout is None:
+                timeout = heartbeat_timeout
+            else:
+                timeout = min(heartbeat_timeout, timeout)
+
         try:
             sock = self.channel.connection.sock
         except AttributeError as e:
             # Level is set to debug because otherwise we would spam the logs
-            LOG.debug('Failed to get socket attribute: %s' % str(e))
-            sock = None
-
-        if sock:
-            orig_timeout = sock.gettimeout()
+            LOG.debug('[%s] Failed to get socket attribute: %s'
+                      % (self.connection_id, str(e)))
+        else:
             sock.settimeout(timeout)
+            # TCP_USER_TIMEOUT is not defined on Windows and Mac OS X
+            if sys.platform != 'win32' and sys.platform != 'darwin':
+                try:
+                    timeout = timeout * 1000 if timeout is not None else 0
+                    # NOTE(gdavoian): only integers and strings are allowed
+                    # as socket options' values, and TCP_USER_TIMEOUT option
+                    # can take only integer values, so we round-up the timeout
+                    # to the nearest integer in order to ensure that the
+                    # connection is not broken before the expected timeout
+                    sock.setsockopt(socket.IPPROTO_TCP,
+                                    TCP_USER_TIMEOUT,
+                                    int(math.ceil(timeout)))
+                except socket.error as error:
+                    code = error[0]
+                    # TCP_USER_TIMEOUT not defined on kernels <2.6.37
+                    if code != errno.ENOPROTOOPT:
+                        raise
+
+    @contextlib.contextmanager
+    def _transport_socket_timeout(self, timeout):
+        self.set_transport_socket_timeout(timeout)
         yield
-        if sock:
-            sock.settimeout(orig_timeout)
+        self.set_transport_socket_timeout()
 
     def _heartbeat_check(self):
         # NOTE(sileht): we are supposed to send at least one heartbeat
         # every heartbeat_timeout_threshold, so no need to way more
-        with self._transport_socket_timeout(
-                self.heartbeat_timeout_threshold):
-            self.connection.heartbeat_check(
-                rate=self.heartbeat_rate)
+        self.connection.heartbeat_check(rate=self.heartbeat_rate)
 
     def _heartbeat_start(self):
         if self._heartbeat_supported_and_enabled():
-            self._heartbeat_exit_event = threading.Event()
+            self._heartbeat_exit_event = eventletutils.Event()
             self._heartbeat_thread = threading.Thread(
                 target=self._heartbeat_thread_job)
             self._heartbeat_thread.daemon = True
@@ -865,24 +975,20 @@ class Connection(object):
         while not self._heartbeat_exit_event.is_set():
             with self._connection_lock.for_heartbeat():
 
-                recoverable_errors = (
-                    self.connection.recoverable_channel_errors +
-                    self.connection.recoverable_connection_errors)
-
                 try:
                     try:
                         self._heartbeat_check()
                         # NOTE(sileht): We need to drain event to receive
                         # heartbeat from the broker but don't hold the
                         # connection too much times. In amqpdriver a connection
-                        # is used exclusivly for read or for write, so we have
+                        # is used exclusively for read or for write, so we have
                         # to do this for connection used for write drain_events
                         # already do that for other connection
                         try:
                             self.connection.drain_events(timeout=0.001)
                         except socket.timeout:
                             pass
-                    except recoverable_errors as exc:
+                    except kombu.exceptions.OperationalError as exc:
                         LOG.info(_LI("A recoverable connection/channel error "
                                      "occurred, trying to reconnect: %s"), exc)
                         self.ensure_connection()
@@ -907,8 +1013,13 @@ class Connection(object):
 
         def _declare_consumer():
             consumer.declare(self)
-            self._consumers.append(consumer)
-            self._new_consumers.append(consumer)
+            tag = self._active_tags.get(consumer.queue_name)
+            if tag is None:
+                tag = next(self._tags)
+                self._active_tags[consumer.queue_name] = tag
+                self._new_tags.add(tag)
+
+            self._consumers[consumer] = tag
             return consumer
 
         with self._connection_lock:
@@ -927,7 +1038,7 @@ class Connection(object):
 
         def _recoverable_error_callback(exc):
             if not isinstance(exc, rpc_common.Timeout):
-                self._new_consumers = self._consumers
+                self._new_tags = set(self._consumers.values())
             timer.check_return(_raise_timeout, exc)
 
         def _error_callback(exc):
@@ -943,11 +1054,11 @@ class Connection(object):
             if not self.connection.connected:
                 raise self.connection.recoverable_connection_errors[0]
 
-            if self._new_consumers:
-                for tag, consumer in enumerate(self._consumers):
-                    if consumer in self._new_consumers:
-                        consumer.consume(tag=tag)
-                self._new_consumers = []
+            while self._new_tags:
+                for consumer, tag in self._consumers.items():
+                    if tag in self._new_tags:
+                        consumer.consume(self, tag=tag)
+                        self._new_tags.remove(tag)
 
             poll_timeout = (self._poll_timeout if timeout is None
                             else min(timeout, self._poll_timeout))
@@ -964,6 +1075,12 @@ class Connection(object):
                 except socket.timeout as exc:
                     poll_timeout = timer.check_return(
                         _raise_timeout, exc, maximum=self._poll_timeout)
+                except self.connection.channel_errors as exc:
+                    if exc.code == 406 and exc.method_name == 'Basic.ack':
+                        # NOTE(gordc): occasionally multiple workers will grab
+                        # same message and acknowledge it. if it happens, meh.
+                        raise self.connection.recoverable_channel_errors[0]
+                    raise
 
         with self._connection_lock:
             self.ensure(_consume,
@@ -1042,39 +1159,37 @@ class Connection(object):
         with self._connection_lock:
             self.ensure(method, retry=retry, error_callback=_error_callback)
 
+    def _get_connection_info(self):
+        info = self.connection.info()
+        client_port = None
+        if (self.channel and hasattr(self.channel.connection, 'sock')
+                and self.channel.connection.sock):
+            client_port = self.channel.connection.sock.getsockname()[1]
+        info.update({'client_port': client_port,
+                     'connection_id': self.connection_id})
+        return info
+
     def _publish(self, exchange, msg, routing_key=None, timeout=None):
         """Publish a message."""
-        producer = kombu.messaging.Producer(exchange=exchange,
-                                            channel=self.channel,
-                                            auto_declare=not exchange.passive,
-                                            routing_key=routing_key)
 
-        # NOTE(sileht): no need to wait more, caller expects
-        # a answer before timeout is reached
-        transport_timeout = timeout
-
-        heartbeat_timeout = self.heartbeat_timeout_threshold
-        if (self._heartbeat_supported_and_enabled() and (
-                transport_timeout is None or
-                transport_timeout > heartbeat_timeout)):
-            # NOTE(sileht): we are supposed to send heartbeat every
-            # heartbeat_timeout, no need to wait more otherwise will
-            # disconnect us, so raise timeout earlier ourself
-            transport_timeout = heartbeat_timeout
+        if not (exchange.passive or exchange.name in self._declared_exchanges):
+                exchange(self.channel).declare()
+                self._declared_exchanges.add(exchange.name)
 
         log_info = {'msg': msg,
                     'who': exchange or 'default',
                     'key': routing_key}
         LOG.trace('Connection._publish: sending message %(msg)s to'
                   ' %(who)s with routing key %(key)s', log_info)
-        with self._transport_socket_timeout(transport_timeout):
-            producer.publish(msg, expiration=timeout,
-                             compression=self.kombu_compression)
 
-    # List of notification queue declared on the channel to avoid
-    # unnecessary redeclaration. This list is resetted each time
-    # the connection is resetted in Connection._set_current_channel
-    PUBLISHER_DECLARED_QUEUES = collections.defaultdict(set)
+        # NOTE(sileht): no need to wait more, caller expects
+        # a answer before timeout is reached
+        with self._transport_socket_timeout(timeout):
+            self._producer.publish(msg,
+                                   exchange=exchange,
+                                   routing_key=routing_key,
+                                   expiration=timeout,
+                                   compression=self.kombu_compression)
 
     def _publish_and_creates_default_queue(self, exchange, msg,
                                            routing_key=None, timeout=None):
@@ -1094,8 +1209,7 @@ class Connection(object):
         # NOTE(sileht): We only do it once per reconnection
         # the Connection._set_current_channel() is responsible to clear
         # this cache
-        if (queue_indentifier not in
-                self.PUBLISHER_DECLARED_QUEUES[self.channel]):
+        if queue_indentifier not in self._declared_queues:
             queue = kombu.entity.Queue(
                 channel=self.channel,
                 exchange=exchange,
@@ -1109,7 +1223,7 @@ class Connection(object):
                 'Connection._publish_and_creates_default_queue: '
                 'declare queue %(key)s on %(exchange)s exchange', log_info)
             queue.declare()
-            self.PUBLISHER_DECLARED_QUEUES[self.channel].add(queue_indentifier)
+            self._declared_queues.add(queue_indentifier)
 
         self._publish(exchange, msg, routing_key=routing_key, timeout=timeout)
 
@@ -1203,6 +1317,7 @@ class RabbitDriver(amqpdriver.AMQPDriverBase):
         conf.register_opts(rabbit_opts, group=opt_group)
         conf.register_opts(rpc_amqp.amqp_opts, group=opt_group)
         conf.register_opts(base.base_opts, group=opt_group)
+        conf = rpc_common.ConfigOptsProxy(conf, url, opt_group.name)
 
         self.missing_destination_retry_timeout = (
             conf.oslo_messaging_rabbit.kombu_missing_consumer_retry_timeout)
@@ -1210,8 +1325,13 @@ class RabbitDriver(amqpdriver.AMQPDriverBase):
         self.prefetch_size = (
             conf.oslo_messaging_rabbit.rabbit_qos_prefetch_count)
 
+        # the pool configuration properties
+        max_size = conf.oslo_messaging_rabbit.rpc_conn_pool_size
+        min_size = conf.oslo_messaging_rabbit.conn_pool_min_size
+        ttl = conf.oslo_messaging_rabbit.conn_pool_ttl
+
         connection_pool = pool.ConnectionPool(
-            conf, conf.oslo_messaging_rabbit.rpc_conn_pool_size,
+            conf, max_size, min_size, ttl,
             url, Connection)
 
         super(RabbitDriver, self).__init__(
