@@ -16,12 +16,14 @@
 #    under the License.
 
 import abc
+import argparse
 import logging
 import uuid
 
 from oslo_config import cfg
 from oslo_utils import timeutils
 import six
+from stevedore import extension
 from stevedore import named
 
 from oslo_messaging._i18n import _LE
@@ -39,6 +41,7 @@ _notifier_opts = [
     cfg.StrOpt('transport_url',
                deprecated_name='notification_transport_url',
                deprecated_group='DEFAULT',
+               secret=True,
                help='A URL representing the messaging driver to use for '
                     'notifications. If not set, we fall back to the same '
                     'configuration used for RPC.'),
@@ -51,11 +54,58 @@ _notifier_opts = [
                                       group='DEFAULT')
                 ],
                 help='AMQP topic used for OpenStack notifications.'),
+    cfg.IntOpt('retry', default=-1,
+               help='The maximum number of attempts to re-send a notification '
+                    'message which failed to be delivered due to a '
+                    'recoverable error. 0 - No retry, -1 - indefinite'),
 ]
 
 _LOG = logging.getLogger(__name__)
 
 
+def _send_notification():
+    """Command line tool to send notifications manually."""
+    parser = argparse.ArgumentParser(
+        description='Oslo.messaging notification sending',
+    )
+    parser.add_argument('--config-file',
+                        help='Path to configuration file')
+    parser.add_argument('--transport-url',
+                        help='Transport URL')
+    parser.add_argument('--publisher-id',
+                        help='Publisher ID')
+    parser.add_argument('--event-type',
+                        default="test",
+                        help="Event type")
+    parser.add_argument('--topic',
+                        nargs='*',
+                        help="Topic to send to")
+    parser.add_argument('--priority',
+                        default="info",
+                        choices=("info",
+                                 "audit",
+                                 "warn",
+                                 "error",
+                                 "critical",
+                                 "sample"),
+                        help='Event type')
+    parser.add_argument('--driver',
+                        default="messagingv2",
+                        choices=extension.ExtensionManager(
+                            'oslo.messaging.notify.drivers'
+                        ).names(),
+                        help='Notification driver')
+    parser.add_argument('payload')
+    args = parser.parse_args()
+    conf = cfg.ConfigOpts()
+    conf([],
+         default_config_files=[args.config_file] if args.config_file else None)
+    transport = get_notification_transport(conf, url=args.transport_url)
+    notifier = Notifier(transport, args.publisher_id, topics=args.topic,
+                        driver=args.driver)
+    notifier._notify({}, args.event_type, args.payload, args.priority)
+
+
 @six.add_metaclass(abc.ABCMeta)
 class Driver(object):
     """Base driver for Notifications"""
@@ -80,10 +130,11 @@ class Driver(object):
         :type msg: str
         :param priority: priority of the message
         :type priority: str
-        :param retry: an connection retries configuration
-                      None or -1 means to retry forever
-                      0 means no retry
-                      N means N retries
+        :param retry: connection retries configuration (used by the messaging
+                      driver):
+                      None or -1 means to retry forever.
+                      0 means no retry is attempted.
+                      N means attempt at most N retries.
         :type retry: int
         """
         pass
@@ -91,6 +142,31 @@ class Driver(object):
 
 def get_notification_transport(conf, url=None,
                                allowed_remote_exmods=None, aliases=None):
+    """A factory method for Transport objects for notifications.
+
+    This method should be used for notifications, in case notifications are
+    being sent over a different message bus than normal messaging
+    functionality; for example, using a different driver, or with different
+    access permissions.
+
+    If no transport URL is provided, the URL in the notifications section of
+    the config file will be used.  If that URL is also absent, the same
+    transport as specified in the messaging section will be used.
+
+    If a transport URL is provided, then this function works exactly the same
+    as get_transport.
+
+    :param conf: the user configuration
+    :type conf: cfg.ConfigOpts
+    :param url: a transport URL
+    :type url: str or TransportURL
+    :param allowed_remote_exmods: a list of modules which a client using this
+                                  transport will deserialize remote exceptions
+                                  from
+    :type allowed_remote_exmods: list
+    :param aliases: A map of transport alias to transport name
+    :type aliases: dict
+    """
     conf.register_opts(_notifier_opts,
                        group='oslo_messaging_notifications')
     if url is None:
@@ -132,7 +208,7 @@ class Notifier(object):
         notifier = notifier.Notifier(transport,
                                      'compute.host',
                                      driver='messaging',
-                                     topic='notifications')
+                                     topics=['notifications'])
 
     Notifier objects are relatively expensive to instantiate (mostly the cost
     of loading notification drivers), so it is possible to specialize a given
@@ -143,8 +219,8 @@ class Notifier(object):
     """
 
     def __init__(self, transport, publisher_id=None,
-                 driver=None, topic=None,
-                 serializer=None, retry=None):
+                 driver=None, serializer=None, retry=None,
+                 topics=None):
         """Construct a Notifier object.
 
         :param transport: the transport to use for sending messages
@@ -154,15 +230,16 @@ class Notifier(object):
         :type publisher_id: str
         :param driver: a driver to lookup from oslo_messaging.notify.drivers
         :type driver: str
-        :param topic: the topic which to send messages on
-        :type topic: str
         :param serializer: an optional entity serializer
         :type serializer: Serializer
-        :param retry: an connection retries configuration
-                      None or -1 means to retry forever
-                      0 means no retry
-                      N means N retries
+        :param retry: connection retries configuration (used by the messaging
+                      driver):
+                      None or -1 means to retry forever.
+                      0 means no retry is attempted.
+                      N means attempt at most N retries.
         :type retry: int
+        :param topics: the topics which to send messages on
+        :type topics: list of strings
         """
         conf = transport.conf
         conf.register_opts(_notifier_opts,
@@ -170,13 +247,18 @@ class Notifier(object):
 
         self.transport = transport
         self.publisher_id = publisher_id
-        self.retry = retry
+        if retry is not None:
+            self.retry = retry
+        else:
+            self.retry = conf.oslo_messaging_notifications.retry
 
         self._driver_names = ([driver] if driver is not None else
                               conf.oslo_messaging_notifications.driver)
 
-        self._topics = ([topic] if topic is not None else
-                        conf.oslo_messaging_notifications.topics)
+        if topics is not None:
+            self._topics = topics
+        else:
+            self._topics = conf.oslo_messaging_notifications.topics
         self._serializer = serializer or msg_serializer.NoOpSerializer()
 
         self._driver_mgr = named.NamedExtensionManager(
@@ -202,10 +284,11 @@ class Notifier(object):
         :param publisher_id: field in notifications sent, for example
                              'compute.host1'
         :type publisher_id: str
-        :param retry: an connection retries configuration
-                      None or -1 means to retry forever
-                      0 means no retry
-                      N means N retries
+        :param retry: connection retries configuration (used by the messaging
+                      driver):
+                      None or -1 means to retry forever.
+                      0 means no retry is attempted.
+                      N means attempt at most N retries.
         :type retry: int
         """
         return _SubNotifier._prepare(self, publisher_id, retry=retry)
@@ -339,6 +422,14 @@ class Notifier(object):
         """
         self._notify(ctxt, event_type, payload, 'SAMPLE')
 
+    def is_enabled(self):
+        """Check if the notifier will emit notifications anywhere.
+
+        :return: false if the driver of the notifier is set only to noop, true
+                 otherwise
+        """
+        return self._driver_mgr.names() != ['noop']
+
 
 class _SubNotifier(Notifier):
 
