@@ -12,7 +12,6 @@
 #    under the License.
 
 import os
-import threading
 import time
 import uuid
 
@@ -21,6 +20,8 @@ from oslo_config import cfg
 from six import moves
 
 import oslo_messaging
+from oslo_messaging._drivers import kafka_options
+from oslo_messaging._drivers.zmq_driver import zmq_options
 from oslo_messaging.notify import notifier
 from oslo_messaging.tests import utils as test_utils
 
@@ -53,12 +54,13 @@ class TestServerEndpoint(object):
 class TransportFixture(fixtures.Fixture):
     """Fixture defined to setup the oslo_messaging transport."""
 
-    def __init__(self, url):
+    def __init__(self, conf, url):
+        self.conf = conf
         self.url = url
 
     def setUp(self):
         super(TransportFixture, self).setUp()
-        self.transport = oslo_messaging.get_transport(cfg.CONF, url=self.url)
+        self.transport = oslo_messaging.get_transport(self.conf, url=self.url)
 
     def cleanUp(self):
         try:
@@ -75,9 +77,10 @@ class TransportFixture(fixtures.Fixture):
 class RpcServerFixture(fixtures.Fixture):
     """Fixture to setup the TestServerEndpoint."""
 
-    def __init__(self, url, target, endpoint=None, ctrl_target=None,
-                 executor='blocking'):
+    def __init__(self, conf, url, target, endpoint=None, ctrl_target=None,
+                 executor='eventlet'):
         super(RpcServerFixture, self).__init__()
+        self.conf = conf
         self.url = url
         self.target = target
         self.endpoint = endpoint or TestServerEndpoint()
@@ -88,7 +91,7 @@ class RpcServerFixture(fixtures.Fixture):
     def setUp(self):
         super(RpcServerFixture, self).setUp()
         endpoints = [self.endpoint, self]
-        transport = self.useFixture(TransportFixture(self.url))
+        transport = self.useFixture(TransportFixture(self.conf, self.url))
         self.server = oslo_messaging.get_rpc_server(
             transport=transport.transport,
             target=self.target,
@@ -104,15 +107,14 @@ class RpcServerFixture(fixtures.Fixture):
         super(RpcServerFixture, self).cleanUp()
 
     def _start(self):
-        self.thread = threading.Thread(target=self.server.start)
-        self.thread.daemon = True
+        self.thread = test_utils.ServerThreadHelper(self.server)
         self.thread.start()
 
     def _stop(self):
-        self.server.stop()
-        self._ctrl.cast({}, 'ping')
-        self.server.wait()
-        self.thread.join()
+        self.thread.stop()
+        self.thread.join(timeout=30)
+        if self.thread.isAlive():
+            raise Exception("Server did not shutdown correctly")
 
     def ping(self, ctxt):
         pass
@@ -122,17 +124,19 @@ class RpcServerFixture(fixtures.Fixture):
 
 
 class RpcServerGroupFixture(fixtures.Fixture):
-    def __init__(self, url, topic=None, names=None, exchange=None,
-                 use_fanout_ctrl=False):
+    def __init__(self, conf, url, topic=None, names=None, exchange=None,
+                 use_fanout_ctrl=False, endpoint=None):
+        self.conf = conf
         self.url = url
-        # NOTE(sileht): topic and servier_name must be uniq
+        # NOTE(sileht): topic and server_name must be unique
         # to be able to run all tests in parallel
         self.topic = topic or str(uuid.uuid4())
-        self.names = names or ["server_%i_%s" % (i, uuid.uuid4())
+        self.names = names or ["server_%i_%s" % (i, str(uuid.uuid4())[:8])
                                for i in range(3)]
         self.exchange = exchange
         self.targets = [self._target(server=n) for n in self.names]
         self.use_fanout_ctrl = use_fanout_ctrl
+        self.endpoint = endpoint
 
     def setUp(self):
         super(RpcServerGroupFixture, self).setUp()
@@ -148,7 +152,9 @@ class RpcServerGroupFixture(fixtures.Fixture):
         ctrl = None
         if self.use_fanout_ctrl:
             ctrl = self._target(fanout=True)
-        server = RpcServerFixture(self.url, target, ctrl_target=ctrl)
+        server = RpcServerFixture(self.conf, self.url, target,
+                                  endpoint=self.endpoint,
+                                  ctrl_target=ctrl)
         return server
 
     def client(self, server=None, cast=False):
@@ -157,12 +163,12 @@ class RpcServerGroupFixture(fixtures.Fixture):
         else:
             if server == 'all':
                 target = self._target(fanout=True)
-            elif server >= 0 and server < len(self.targets):
+            elif 0 <= server < len(self.targets):
                 target = self.targets[server]
             else:
                 raise ValueError("Invalid value for server: %r" % server)
 
-        transport = self.useFixture(TransportFixture(self.url))
+        transport = self.useFixture(TransportFixture(self.conf, self.url))
         client = ClientStub(transport.transport, target, cast=cast,
                             timeout=5)
         transport.wait()
@@ -176,7 +182,7 @@ class RpcServerGroupFixture(fixtures.Fixture):
             if server == 'all':
                 for s in self.servers:
                     s.syncq.get(timeout=5)
-            elif server >= 0 and server < len(self.targets):
+            elif 0 <= server < len(self.targets):
                 self.servers[server].syncq.get(timeout=5)
             else:
                 raise ValueError("Invalid value for server: %r" % server)
@@ -276,40 +282,72 @@ class IsValidDistributionOf(object):
 class SkipIfNoTransportURL(test_utils.BaseTestCase):
     def setUp(self, conf=cfg.CONF):
         super(SkipIfNoTransportURL, self).setUp(conf=conf)
-        self.url = os.environ.get('TRANSPORT_URL')
+
+        driver = os.environ.get("TRANSPORT_DRIVER")
+        if driver:
+            self.url = os.environ.get('PIFPAF_URL')
+            if driver == "pika" and self.url:
+                self.url = self.url.replace("rabbit://", "pika://")
+        else:
+            self.url = os.environ.get('TRANSPORT_URL')
+
         if not self.url:
             self.skipTest("No transport url configured")
 
+        transport_url = oslo_messaging.TransportURL.parse(conf, self.url)
+
+        zmq_options.register_opts(conf, transport_url)
+
         zmq_matchmaker = os.environ.get('ZMQ_MATCHMAKER')
         if zmq_matchmaker:
-            self.config(rpc_zmq_matchmaker=zmq_matchmaker)
+            self.config(rpc_zmq_matchmaker=zmq_matchmaker,
+                        group="oslo_messaging_zmq")
         zmq_ipc_dir = os.environ.get('ZMQ_IPC_DIR')
         if zmq_ipc_dir:
-            self.config(rpc_zmq_ipc_dir=zmq_ipc_dir)
+            self.config(group="oslo_messaging_zmq",
+                        rpc_zmq_ipc_dir=zmq_ipc_dir)
         zmq_redis_port = os.environ.get('ZMQ_REDIS_PORT')
         if zmq_redis_port:
-            self.config(port=zmq_redis_port, group="matchmaker_redis")
+            self.config(port=zmq_redis_port,
+                        check_timeout=10000,
+                        wait_timeout=1000,
+                        group="matchmaker_redis")
+        zmq_use_pub_sub = os.environ.get('ZMQ_USE_PUB_SUB')
+        zmq_use_router_proxy = os.environ.get('ZMQ_USE_ROUTER_PROXY')
+        zmq_use_acks = os.environ.get('ZMQ_USE_ACKS')
+        self.config(use_pub_sub=zmq_use_pub_sub,
+                    use_router_proxy=zmq_use_router_proxy,
+                    rpc_use_acks=zmq_use_acks,
+                    group='oslo_messaging_zmq')
+        zmq_use_dynamic_connections = \
+            os.environ.get('ZMQ_USE_DYNAMIC_CONNECTIONS')
+        self.config(use_dynamic_connections=zmq_use_dynamic_connections,
+                    group='oslo_messaging_zmq')
+
+        kafka_options.register_opts(conf, transport_url)
+
+        self.config(producer_batch_size=0,
+                    group='oslo_messaging_kafka')
 
 
 class NotificationFixture(fixtures.Fixture):
-    def __init__(self, url, topics):
+    def __init__(self, conf, url, topics, batch=None):
         super(NotificationFixture, self).__init__()
+        self.conf = conf
         self.url = url
         self.topics = topics
         self.events = moves.queue.Queue()
         self.name = str(id(self))
+        self.batch = batch
 
     def setUp(self):
         super(NotificationFixture, self).setUp()
         targets = [oslo_messaging.Target(topic=t) for t in self.topics]
         # add a special topic for internal notifications
         targets.append(oslo_messaging.Target(topic=self.name))
-        transport = self.useFixture(TransportFixture(self.url))
-        self.server = oslo_messaging.get_notification_listener(
-            transport.transport,
-            targets,
-            [self])
-        self._ctrl = self.notifier('internal', topic=self.name)
+        transport = self.useFixture(TransportFixture(self.conf, self.url))
+        self.server = self._get_server(transport, targets)
+        self._ctrl = self.notifier('internal', topics=[self.name])
         self._start()
         transport.wait()
 
@@ -317,23 +355,28 @@ class NotificationFixture(fixtures.Fixture):
         self._stop()
         super(NotificationFixture, self).cleanUp()
 
+    def _get_server(self, transport, targets):
+        return oslo_messaging.get_notification_listener(
+            transport.transport,
+            targets,
+            [self], 'eventlet')
+
     def _start(self):
-        self.thread = threading.Thread(target=self.server.start)
-        self.thread.daemon = True
+        self.thread = test_utils.ServerThreadHelper(self.server)
         self.thread.start()
 
     def _stop(self):
-        self.server.stop()
-        self._ctrl.sample({}, 'shutdown', 'shutdown')
-        self.server.wait()
-        self.thread.join()
+        self.thread.stop()
+        self.thread.join(timeout=30)
+        if self.thread.isAlive():
+            raise Exception("Server did not shutdown properly")
 
-    def notifier(self, publisher, topic=None):
-        transport = self.useFixture(TransportFixture(self.url))
+    def notifier(self, publisher, topics=None):
+        transport = self.useFixture(TransportFixture(self.conf, self.url))
         n = notifier.Notifier(transport.transport,
                               publisher,
                               driver='messaging',
-                              topic=topic or self.topics[0])
+                              topics=topics or self.topics)
         transport.wait()
         return n
 
@@ -366,3 +409,39 @@ class NotificationFixture(fixtures.Fixture):
         except moves.queue.Empty:
             pass
         return results
+
+
+class BatchNotificationFixture(NotificationFixture):
+    def __init__(self, conf, url, topics, batch_size=5, batch_timeout=2):
+        super(BatchNotificationFixture, self).__init__(conf, url, topics)
+        self.batch_size = batch_size
+        self.batch_timeout = batch_timeout
+
+    def _get_server(self, transport, targets):
+        return oslo_messaging.get_batch_notification_listener(
+            transport.transport,
+            targets,
+            [self], 'eventlet',
+            batch_timeout=self.batch_timeout,
+            batch_size=self.batch_size)
+
+    def debug(self, messages):
+        self.events.put(['debug', messages])
+
+    def audit(self, messages):
+        self.events.put(['audit', messages])
+
+    def info(self, messages):
+        self.events.put(['info', messages])
+
+    def warn(self, messages):
+        self.events.put(['warn', messages])
+
+    def error(self, messages):
+        self.events.put(['error', messages])
+
+    def critical(self, messages):
+        self.events.put(['critical', messages])
+
+    def sample(self, messages):
+        pass  # Just used for internal shutdown control
