@@ -14,6 +14,7 @@
 #    under the License.
 
 from oslo_config import cfg
+from six.moves import mock
 import testscenarios
 
 import oslo_messaging
@@ -54,7 +55,7 @@ class TestCastCall(test_utils.BaseTestCase):
         transport = _FakeTransport(self.conf)
         client = oslo_messaging.RPCClient(transport, oslo_messaging.Target())
 
-        self.mox.StubOutWithMock(transport, '_send')
+        transport._send = mock.Mock()
 
         msg = dict(method='foo', args=self.args)
         kwargs = {'retry': None}
@@ -62,12 +63,14 @@ class TestCastCall(test_utils.BaseTestCase):
             kwargs['wait_for_reply'] = True
             kwargs['timeout'] = None
 
-        transport._send(oslo_messaging.Target(), self.ctxt, msg, **kwargs)
-        self.mox.ReplayAll()
-
         method = client.call if self.call else client.cast
         method(self.ctxt, 'foo', **self.args)
 
+        transport._send.assert_called_once_with(oslo_messaging.Target(),
+                                                self.ctxt,
+                                                msg,
+                                                **kwargs)
+
 
 class TestCastToTarget(test_utils.BaseTestCase):
 
@@ -122,21 +125,21 @@ class TestCastToTarget(test_utils.BaseTestCase):
               prepare=dict(namespace='testnamespace'),
               expect=dict(namespace='testnamespace'))),
         ('ctor_version',
-         dict(ctor=dict(version='testversion'),
+         dict(ctor=dict(version='1.1'),
               prepare={},
-              expect=dict(version='testversion'))),
+              expect=dict(version='1.1'))),
         ('prepare_version',
          dict(ctor={},
-              prepare=dict(version='testversion'),
-              expect=dict(version='testversion'))),
+              prepare=dict(version='1.1'),
+              expect=dict(version='1.1'))),
         ('prepare_version_none',
-         dict(ctor=dict(version='testversion'),
+         dict(ctor=dict(version='1.1'),
               prepare=dict(version=None),
               expect={})),
         ('both_version',
          dict(ctor=dict(version='ctorversion'),
-              prepare=dict(version='testversion'),
-              expect=dict(version='testversion'))),
+              prepare=dict(version='1.1'),
+              expect=dict(version='1.1'))),
         ('ctor_server',
          dict(ctor=dict(server='testserver'),
               prepare={},
@@ -191,16 +194,13 @@ class TestCastToTarget(test_utils.BaseTestCase):
         transport = _FakeTransport(self.conf)
         client = oslo_messaging.RPCClient(transport, target)
 
-        self.mox.StubOutWithMock(transport, '_send')
+        transport._send = mock.Mock()
 
         msg = dict(method='foo', args={})
         if 'namespace' in self.expect:
             msg['namespace'] = self.expect['namespace']
         if 'version' in self.expect:
             msg['version'] = self.expect['version']
-        transport._send(expect_target, {}, msg, retry=None)
-
-        self.mox.ReplayAll()
 
         if self.prepare:
             client = client.prepare(**self.prepare)
@@ -208,6 +208,11 @@ class TestCastToTarget(test_utils.BaseTestCase):
                 client = client.prepare(**self.prepare)
         client.cast({}, 'foo')
 
+        transport._send.assert_called_once_with(expect_target,
+                                                {},
+                                                msg,
+                                                retry=None)
+
 
 TestCastToTarget.generate_scenarios()
 
@@ -221,7 +226,7 @@ class TestCallTimeout(test_utils.BaseTestCase):
         ('all_none',
          dict(confval=None, ctor=None, prepare=_notset, expect=None)),
         ('confval',
-         dict(confval=21.1, ctor=None, prepare=_notset, expect=21.1)),
+         dict(confval=21, ctor=None, prepare=_notset, expect=21)),
         ('ctor',
          dict(confval=None, ctor=21.1, prepare=_notset, expect=21.1)),
         ('ctor_zero',
@@ -241,18 +246,20 @@ class TestCallTimeout(test_utils.BaseTestCase):
         client = oslo_messaging.RPCClient(transport, oslo_messaging.Target(),
                                           timeout=self.ctor)
 
-        self.mox.StubOutWithMock(transport, '_send')
+        transport._send = mock.Mock()
 
         msg = dict(method='foo', args={})
         kwargs = dict(wait_for_reply=True, timeout=self.expect, retry=None)
-        transport._send(oslo_messaging.Target(), {}, msg, **kwargs)
-
-        self.mox.ReplayAll()
 
         if self.prepare is not _notset:
             client = client.prepare(timeout=self.prepare)
         client.call({}, 'foo')
 
+        transport._send.assert_called_once_with(oslo_messaging.Target(),
+                                                {},
+                                                msg,
+                                                **kwargs)
+
 
 class TestCallRetry(test_utils.BaseTestCase):
 
@@ -270,19 +277,21 @@ class TestCallRetry(test_utils.BaseTestCase):
         client = oslo_messaging.RPCClient(transport, oslo_messaging.Target(),
                                           retry=self.ctor)
 
-        self.mox.StubOutWithMock(transport, '_send')
+        transport._send = mock.Mock()
 
         msg = dict(method='foo', args={})
         kwargs = dict(wait_for_reply=True, timeout=60,
                       retry=self.expect)
-        transport._send(oslo_messaging.Target(), {}, msg, **kwargs)
-
-        self.mox.ReplayAll()
 
         if self.prepare is not _notset:
             client = client.prepare(retry=self.prepare)
         client.call({}, 'foo')
 
+        transport._send.assert_called_once_with(oslo_messaging.Target(),
+                                                {},
+                                                msg,
+                                                **kwargs)
+
 
 class TestCallFanout(test_utils.BaseTestCase):
 
@@ -328,37 +337,47 @@ class TestSerializer(test_utils.BaseTestCase):
         client = oslo_messaging.RPCClient(transport, oslo_messaging.Target(),
                                           serializer=serializer)
 
-        self.mox.StubOutWithMock(transport, '_send')
+        transport._send = mock.Mock()
 
-        msg = dict(method='foo',
-                   args=dict([(k, 's' + v) for k, v in self.args.items()]))
         kwargs = dict(wait_for_reply=True, timeout=None) if self.call else {}
         kwargs['retry'] = None
-        transport._send(oslo_messaging.Target(),
-                        dict(user='alice'),
-                        msg,
-                        **kwargs).AndReturn(self.retval)
 
-        self.mox.StubOutWithMock(serializer, 'serialize_entity')
-        self.mox.StubOutWithMock(serializer, 'deserialize_entity')
-        self.mox.StubOutWithMock(serializer, 'serialize_context')
+        transport._send.return_value = self.retval
 
-        for arg in self.args:
-            serializer.serialize_entity(self.ctxt, arg).AndReturn('s' + arg)
+        serializer.serialize_entity = mock.Mock()
+        serializer.deserialize_entity = mock.Mock()
+        serializer.serialize_context = mock.Mock()
+
+        msg = dict(method='foo', args=dict())
+        expected_side_effect = []
+        for k, v in self.args.items():
+            msg['args'][k] = 's' + v
+            expected_side_effect.append('s' + v)
+
+        serializer.serialize_entity.side_effect = expected_side_effect
 
         if self.call:
-            serializer.deserialize_entity(self.ctxt, self.retval).\
-                AndReturn('d' + self.retval)
+            serializer.deserialize_entity.return_value = 'd' + self.retval
 
-        serializer.serialize_context(self.ctxt).AndReturn(dict(user='alice'))
-
-        self.mox.ReplayAll()
+        serializer.serialize_context.return_value = dict(user='alice')
 
         method = client.call if self.call else client.cast
         retval = method(self.ctxt, 'foo', **self.args)
         if self.retval is not None:
             self.assertEqual('d' + self.retval, retval)
 
+        transport._send.assert_called_once_with(oslo_messaging.Target(),
+                                                dict(user='alice'),
+                                                msg,
+                                                **kwargs)
+        expected_calls = [mock.call(self.ctxt, arg) for arg in self.args]
+        self.assertEqual(expected_calls,
+                         serializer.serialize_entity.mock_calls)
+        if self.call:
+            serializer.deserialize_entity.assert_called_once_with(self.ctxt,
+                                                                  self.retval)
+        serializer.serialize_context.assert_called_once_with(self.ctxt)
+
 
 class TestVersionCap(test_utils.BaseTestCase):
 
@@ -392,6 +411,14 @@ class TestVersionCap(test_utils.BaseTestCase):
          dict(cap='2.0', prepare_cap=_notset,
               version=None, prepare_version='1.0',
               success=False)),
+        ('ctor_cap_none_version_ok',
+         dict(cap=None, prepare_cap=_notset,
+              version='1.0', prepare_version=_notset,
+              success=True)),
+        ('ctor_cap_version_none_fail',
+         dict(cap='1.0', prepare_cap=_notset,
+              version=None, prepare_version=_notset,
+              success=False)),
     ]
 
     @classmethod
@@ -410,7 +437,7 @@ class TestVersionCap(test_utils.BaseTestCase):
                                           version_cap=self.cap)
 
         if self.success:
-            self.mox.StubOutWithMock(transport, '_send')
+            transport._send = mock.Mock()
 
             if self.prepare_version is not _notset:
                 target = target(version=self.prepare_version)
@@ -424,10 +451,6 @@ class TestVersionCap(test_utils.BaseTestCase):
                 kwargs['wait_for_reply'] = True
                 kwargs['timeout'] = None
 
-            transport._send(target, {}, msg, **kwargs)
-
-            self.mox.ReplayAll()
-
         prep_kwargs = {}
         if self.prepare_cap is not _notset:
             prep_kwargs['version_cap'] = self.prepare_cap
@@ -444,7 +467,7 @@ class TestVersionCap(test_utils.BaseTestCase):
             self.assertFalse(self.success)
         else:
             self.assertTrue(self.success)
-
+            transport._send.assert_called_once_with(target, {}, msg, **kwargs)
 
 TestVersionCap.generate_scenarios()
 
@@ -492,6 +515,21 @@ class TestCanSendVersion(test_utils.BaseTestCase):
               version=None, prepare_version='1.0',
               can_send_version=_notset,
               can_send=False)),
+        ('ctor_cap_none_version_ok',
+         dict(cap=None, prepare_cap=_notset,
+              version='1.0', prepare_version=_notset,
+              can_send_version=_notset,
+              can_send=True)),
+        ('ctor_cap_version_none_fail',
+         dict(cap='1.0', prepare_cap=_notset,
+              version=None, prepare_version=_notset,
+              can_send_version=_notset,
+              can_send=False)),
+        ('ctor_cap_version_can_send_none_fail',
+         dict(cap='1.0', prepare_cap=_notset,
+              version='1.0', prepare_version=_notset,
+              can_send_version=None,
+              can_send=False)),
     ]
 
     def test_version_cap(self):
@@ -513,7 +551,21 @@ class TestCanSendVersion(test_utils.BaseTestCase):
 
         if self.can_send_version is not _notset:
             can_send = client.can_send_version(version=self.can_send_version)
+            call_context_can_send = client.prepare().can_send_version(
+                version=self.can_send_version)
+            self.assertEqual(can_send, call_context_can_send)
         else:
             can_send = client.can_send_version()
 
         self.assertEqual(self.can_send, can_send)
+
+    def test_invalid_version_type(self):
+        target = oslo_messaging.Target(topic='sometopic')
+        transport = _FakeTransport(self.conf)
+        client = oslo_messaging.RPCClient(transport, target)
+        self.assertRaises(exceptions.MessagingException,
+                          client.prepare, version='5')
+        self.assertRaises(exceptions.MessagingException,
+                          client.prepare, version='5.a')
+        self.assertRaises(exceptions.MessagingException,
+                          client.prepare, version='5.5.a')
