@@ -13,14 +13,19 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-import eventlet
 import threading
+import warnings
 
+import eventlet
+import fixtures
 from oslo_config import cfg
+from six.moves import mock
 import testscenarios
 
-import mock
 import oslo_messaging
+from oslo_messaging import rpc
+from oslo_messaging.rpc import dispatcher
+from oslo_messaging.rpc import server as rpc_server_module
 from oslo_messaging import server as server_module
 from oslo_messaging.tests import utils as test_utils
 
@@ -29,7 +34,7 @@ load_tests = testscenarios.load_tests_apply_scenarios
 
 class ServerSetupMixin(object):
 
-    class Server(threading.Thread):
+    class Server(object):
         def __init__(self, transport, topic, server, endpoint, serializer):
             self.controller = ServerSetupMixin.ServerController()
             target = oslo_messaging.Target(topic=topic, server=server)
@@ -39,9 +44,6 @@ class ServerSetupMixin(object):
                                                          self.controller],
                                                         serializer=serializer)
 
-            super(ServerSetupMixin.Server, self).__init__()
-            self.daemon = True
-
         def wait(self):
             # Wait for the executor to process the stop message, indicating all
             # test messages have been processed
@@ -52,7 +54,7 @@ class ServerSetupMixin(object):
             self.server.stop()
             self.server.wait()
 
-        def run(self):
+        def start(self):
             self.server.start()
 
     class ServerController(object):
@@ -86,10 +88,7 @@ class ServerSetupMixin(object):
                              endpoint=endpoint,
                              serializer=self.serializer)
 
-        thread = threading.Thread(target=server.start)
-        thread.daemon = True
-        thread.start()
-
+        server.start()
         return server
 
     def _stop_server(self, client, server, topic=None):
@@ -118,16 +117,35 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
         target = oslo_messaging.Target(topic='foo', server='bar')
         endpoints = [object()]
         serializer = object()
+        access_policy = dispatcher.DefaultRPCAccessPolicy
 
-        server = oslo_messaging.get_rpc_server(transport, target, endpoints,
-                                               serializer=serializer)
-
+        with warnings.catch_warnings(record=True) as capture:
+            warnings.simplefilter("always", FutureWarning)
+            server = oslo_messaging.get_rpc_server(transport,
+                                                   target,
+                                                   endpoints,
+                                                   serializer=serializer,
+                                                   access_policy=access_policy)
+        self.assertEqual(0, len(capture))
         self.assertIs(server.conf, self.conf)
         self.assertIs(server.transport, transport)
         self.assertIsInstance(server.dispatcher, oslo_messaging.RPCDispatcher)
         self.assertIs(server.dispatcher.endpoints, endpoints)
         self.assertIs(server.dispatcher.serializer, serializer)
-        self.assertEqual('blocking', server.executor)
+        self.assertEqual('blocking', server.executor_type)
+
+    def test_constructor_without_explicit_RPCAccessPolicy(self):
+        transport = oslo_messaging.get_transport(self.conf, url='fake:')
+        target = oslo_messaging.Target(topic='foo', server='bar')
+        endpoints = [object()]
+        serializer = object()
+        with warnings.catch_warnings(record=True) as capture:
+            warnings.simplefilter("always", FutureWarning)
+            oslo_messaging.get_rpc_server(transport, target,
+                                          endpoints, serializer=serializer)
+        self.assertEqual(1, len(capture))
+        w = capture[0]
+        self.assertEqual(FutureWarning, w.category)
 
     def test_server_wait_method(self):
         transport = oslo_messaging.get_transport(self.conf, url='fake:')
@@ -136,11 +154,11 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
         serializer = object()
 
         class MagicMockIgnoreArgs(mock.MagicMock):
-            '''MagicMock ignores arguments.
+            """MagicMock ignores arguments.
 
             A MagicMock which can never misinterpret the arguments passed to
             it during construction.
-            '''
+            """
 
             def __init__(self, *args, **kwargs):
                 super(MagicMockIgnoreArgs, self).__init__()
@@ -149,15 +167,16 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
                                                serializer=serializer)
         # Mocking executor
         server._executor_cls = MagicMockIgnoreArgs
+        server._create_listener = MagicMockIgnoreArgs()
+        server.dispatcher = MagicMockIgnoreArgs()
         # Here assigning executor's listener object to listener variable
         # before calling wait method, because in wait method we are
         # setting executor to None.
         server.start()
-        listener = server._executor_obj.listener
+        listener = server.listener
         server.stop()
         # call server wait method
         server.wait()
-        self.assertIsNone(server._executor_obj)
         self.assertEqual(1, listener.cleanup.call_count)
 
     def test_no_target_server(self):
@@ -331,6 +350,24 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
             def ping(self, ctxt, arg):
                 raise ValueError(arg)
 
+        debugs = []
+        errors = []
+
+        def stub_debug(msg, *a, **kw):
+            if (a and len(a) == 1 and isinstance(a[0], dict) and a[0]):
+                a = a[0]
+            debugs.append(str(msg) % a)
+
+        def stub_error(msg, *a, **kw):
+            if (a and len(a) == 1 and isinstance(a[0], dict) and a[0]):
+                a = a[0]
+            errors.append(str(msg) % a)
+
+        self.useFixture(fixtures.MockPatchObject(
+            rpc_server_module.LOG, 'debug', stub_debug))
+        self.useFixture(fixtures.MockPatchObject(
+            rpc_server_module.LOG, 'error', stub_error))
+
         server_thread = self._setup_server(transport, TestEndpoint())
         client = self._setup_client(transport)
 
@@ -339,6 +376,8 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
         except Exception as ex:
             self.assertIsInstance(ex, ValueError)
             self.assertEqual('dsfoo', str(ex))
+            self.assertTrue(len(debugs) == 0)
+            self.assertGreater(len(errors), 0)
         else:
             self.assertTrue(False)
 
@@ -347,6 +386,24 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
     def test_expected_failure(self):
         transport = oslo_messaging.get_transport(self.conf, url='fake:')
 
+        debugs = []
+        errors = []
+
+        def stub_debug(msg, *a, **kw):
+            if (a and len(a) == 1 and isinstance(a[0], dict) and a[0]):
+                a = a[0]
+            debugs.append(str(msg) % a)
+
+        def stub_error(msg, *a, **kw):
+            if (a and len(a) == 1 and isinstance(a[0], dict) and a[0]):
+                a = a[0]
+            errors.append(str(msg) % a)
+
+        self.useFixture(fixtures.MockPatchObject(
+            rpc_server_module.LOG, 'debug', stub_debug))
+        self.useFixture(fixtures.MockPatchObject(
+            rpc_server_module.LOG, 'error', stub_error))
+
         class TestEndpoint(object):
             @oslo_messaging.expected_exceptions(ValueError)
             def ping(self, ctxt, arg):
@@ -360,6 +417,8 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
         except Exception as ex:
             self.assertIsInstance(ex, ValueError)
             self.assertEqual('dsfoo', str(ex))
+            self.assertGreater(len(debugs), 0)
+            self.assertTrue(len(errors) == 0)
         else:
             self.assertTrue(False)
 
@@ -455,9 +514,9 @@ class TestMultipleServers(test_utils.BaseTestCase, ServerSetupMixin):
             return scenario
 
         for f in [filter_fanout_call, filter_same_topic_and_server]:
-            cls.scenarios = filter(f, cls.scenarios)
+            cls.scenarios = [i for i in cls.scenarios if f(i)]
         for m in [fanout_to_servers, single_topic_multi_endpoints]:
-            cls.scenarios = map(m, cls.scenarios)
+            cls.scenarios = [m(i) for i in cls.scenarios]
 
     def __init__(self, *args):
         super(TestMultipleServers, self).__init__(*args)
@@ -491,9 +550,9 @@ class TestMultipleServers(test_utils.BaseTestCase, ServerSetupMixin):
         else:
             endpoint1 = endpoint2 = TestEndpoint()
 
-        thread1 = self._setup_server(transport1, endpoint1,
+        server1 = self._setup_server(transport1, endpoint1,
                                      topic=self.topic1, server=self.server1)
-        thread2 = self._setup_server(transport2, endpoint2,
+        server2 = self._setup_server(transport2, endpoint2,
                                      topic=self.topic2, server=self.server2)
 
         client1 = self._setup_client(transport1, topic=self.topic1)
@@ -512,12 +571,10 @@ class TestMultipleServers(test_utils.BaseTestCase, ServerSetupMixin):
         (client1.call if self.call1 else client1.cast)({}, 'ping', arg='1')
         (client2.call if self.call2 else client2.cast)({}, 'ping', arg='2')
 
-        self.assertTrue(thread1.isAlive())
         self._stop_server(client1.prepare(fanout=None),
-                          thread1, topic=self.topic1)
-        self.assertTrue(thread2.isAlive())
+                          server1, topic=self.topic1)
         self._stop_server(client2.prepare(fanout=None),
-                          thread2, topic=self.topic2)
+                          server2, topic=self.topic2)
 
         def check(pings, expect):
             self.assertEqual(len(expect), len(pings))
@@ -539,7 +596,7 @@ class TestServerLocking(test_utils.BaseTestCase):
         super(TestServerLocking, self).setUp(conf=cfg.ConfigOpts())
 
         def _logmethod(name):
-            def method(self):
+            def method(self, *args, **kwargs):
                 with self._lock:
                     self._calls.append(name)
             return method
@@ -550,18 +607,21 @@ class TestServerLocking(test_utils.BaseTestCase):
             def __init__(self, *args, **kwargs):
                 self._lock = threading.Lock()
                 self._calls = []
-                self.listener = mock.MagicMock()
                 executors.append(self)
 
-            start = _logmethod('start')
-            stop = _logmethod('stop')
-            wait = _logmethod('wait')
-            execute = _logmethod('execute')
+            submit = _logmethod('submit')
+            shutdown = _logmethod('shutdown')
 
         self.executors = executors
 
-        self.server = oslo_messaging.MessageHandlingServer(mock.Mock(),
-                                                           mock.Mock())
+        class MessageHandlingServerImpl(oslo_messaging.MessageHandlingServer):
+            def _create_listener(self):
+                return mock.Mock()
+
+            def _process_incoming(self, incoming):
+                pass
+
+        self.server = MessageHandlingServerImpl(mock.Mock(), mock.Mock())
         self.server._executor_cls = FakeExecutor
 
     def test_start_stop_wait(self):
@@ -571,11 +631,9 @@ class TestServerLocking(test_utils.BaseTestCase):
         self.server.stop()
         self.server.wait()
 
-        self.assertEqual(len(self.executors), 1)
-        executor = self.executors[0]
-        self.assertEqual(executor._calls,
-                         ['start', 'execute', 'stop', 'wait'])
-        self.assertTrue(executor.listener.cleanup.called)
+        self.assertEqual(1, len(self.executors))
+        self.assertEqual(['shutdown'], self.executors[0]._calls)
+        self.assertTrue(self.server.listener.cleanup.called)
 
     def test_reversed_order(self):
         # Test that if we call wait, stop, start, these will be correctly
@@ -593,10 +651,8 @@ class TestServerLocking(test_utils.BaseTestCase):
 
         self.server.wait()
 
-        self.assertEqual(len(self.executors), 1)
-        executor = self.executors[0]
-        self.assertEqual(executor._calls,
-                         ['start', 'execute', 'stop', 'wait'])
+        self.assertEqual(1, len(self.executors))
+        self.assertEqual(['shutdown'], self.executors[0]._calls)
 
     def test_wait_for_running_task(self):
         # Test that if 2 threads call a method simultaneously, both will wait,
@@ -608,16 +664,16 @@ class TestServerLocking(test_utils.BaseTestCase):
         running_event = threading.Event()
         done_event = threading.Event()
 
-        runner = [None]
+        _runner = [None]
 
         class SteppingFakeExecutor(self.server._executor_cls):
-            def start(self):
+            def __init__(self, *args, **kwargs):
                 # Tell the test which thread won the race
-                runner[0] = eventlet.getcurrent()
+                _runner[0] = eventlet.getcurrent()
                 running_event.set()
 
                 start_event.wait()
-                super(SteppingFakeExecutor, self).start()
+                super(SteppingFakeExecutor, self).__init__(*args, **kwargs)
                 done_event.set()
 
                 finish_event.wait()
@@ -629,7 +685,7 @@ class TestServerLocking(test_utils.BaseTestCase):
 
         # Wait until one of the threads starts running
         running_event.wait()
-        runner = runner[0]
+        runner = _runner[0]
         waiter = start2 if runner == start1 else start2
 
         waiter_finished = threading.Event()
@@ -637,18 +693,16 @@ class TestServerLocking(test_utils.BaseTestCase):
 
         # At this point, runner is running start(), and waiter() is waiting for
         # it to complete. runner has not yet logged anything.
-        self.assertEqual(1, len(self.executors))
-        executor = self.executors[0]
-
-        self.assertEqual(executor._calls, [])
+        self.assertEqual(0, len(self.executors))
         self.assertFalse(waiter_finished.is_set())
 
         # Let the runner log the call
         start_event.set()
         done_event.wait()
 
-        # We haven't signalled completion yet, so execute shouldn't have run
-        self.assertEqual(executor._calls, ['start'])
+        # We haven't signalled completion yet, so submit shouldn't have run
+        self.assertEqual(1, len(self.executors))
+        self.assertEqual([], self.executors[0]._calls)
         self.assertFalse(waiter_finished.is_set())
 
         # Let the runner complete
@@ -659,7 +713,8 @@ class TestServerLocking(test_utils.BaseTestCase):
         # Check that both threads have finished, start was only called once,
         # and execute ran
         self.assertTrue(waiter_finished.is_set())
-        self.assertEqual(executor._calls, ['start', 'execute'])
+        self.assertEqual(1, len(self.executors))
+        self.assertEqual([], self.executors[0]._calls)
 
     def test_start_stop_wait_stop_wait(self):
         # Test that we behave correctly when calling stop/wait more than once.
@@ -672,10 +727,8 @@ class TestServerLocking(test_utils.BaseTestCase):
         self.server.wait()
 
         self.assertEqual(len(self.executors), 1)
-        executor = self.executors[0]
-        self.assertEqual(executor._calls,
-                         ['start', 'execute', 'stop', 'wait'])
-        self.assertTrue(executor.listener.cleanup.called)
+        self.assertEqual(['shutdown'], self.executors[0]._calls)
+        self.assertTrue(self.server.listener.cleanup.called)
 
     def test_state_wrapping(self):
         # Test that we behave correctly if a thread waits, and the server state
@@ -710,7 +763,7 @@ class TestServerLocking(test_utils.BaseTestCase):
 
         # The server should have started, but stop should not have been called
         self.assertEqual(1, len(self.executors))
-        self.assertEqual(self.executors[0]._calls, ['start', 'execute'])
+        self.assertEqual([], self.executors[0]._calls)
         self.assertFalse(thread1_finished.is_set())
 
         self.server.stop()
@@ -719,18 +772,16 @@ class TestServerLocking(test_utils.BaseTestCase):
         # We should have gone through all the states, and thread1 should still
         # be waiting
         self.assertEqual(1, len(self.executors))
-        self.assertEqual(self.executors[0]._calls, ['start', 'execute',
-                                                    'stop', 'wait'])
+        self.assertEqual(['shutdown'], self.executors[0]._calls)
         self.assertFalse(thread1_finished.is_set())
 
         # Start again
         self.server.start()
 
-        # We should now record 2 executors
+        # We should now record 4 executors (2 for each server)
         self.assertEqual(2, len(self.executors))
-        self.assertEqual(self.executors[0]._calls, ['start', 'execute',
-                                                    'stop', 'wait'])
-        self.assertEqual(self.executors[1]._calls, ['start', 'execute'])
+        self.assertEqual(['shutdown'], self.executors[0]._calls)
+        self.assertEqual([], self.executors[1]._calls)
         self.assertFalse(thread1_finished.is_set())
 
         # Allow thread1 to complete
@@ -740,9 +791,8 @@ class TestServerLocking(test_utils.BaseTestCase):
         # thread1 should now have finished, and stop should not have been
         # called again on either the first or second executor
         self.assertEqual(2, len(self.executors))
-        self.assertEqual(self.executors[0]._calls, ['start', 'execute',
-                                                    'stop', 'wait'])
-        self.assertEqual(self.executors[1]._calls, ['start', 'execute'])
+        self.assertEqual(['shutdown'], self.executors[0]._calls)
+        self.assertEqual([], self.executors[1]._calls)
         self.assertTrue(thread1_finished.is_set())
 
     @mock.patch.object(server_module, 'DEFAULT_LOG_AFTER', 1)
@@ -807,24 +857,24 @@ class TestServerLocking(test_utils.BaseTestCase):
 
         # Start the server, which will also instantiate an executor
         self.server.start()
-
-        stop_called = threading.Event()
+        self.server.stop()
+        shutdown_called = threading.Event()
 
         # Patch the executor's stop method to be very slow
-        def slow_stop():
-            stop_called.set()
+        def slow_shutdown(wait):
+            shutdown_called.set()
             eventlet.sleep(10)
-        self.executors[0].stop = slow_stop
+        self.executors[0].shutdown = slow_shutdown
 
-        # Call stop in a new thread
-        thread = eventlet.spawn(self.server.stop)
+        # Call wait in a new thread
+        thread = eventlet.spawn(self.server.wait)
 
         # Wait until the thread is in the slow stop method
-        stop_called.wait()
+        shutdown_called.wait()
 
-        # Call stop again in the main thread with a timeout
+        # Call wait again in the main thread with a timeout
         self.assertRaises(server_module.TaskTimeout,
-                          self.server.stop, timeout=1)
+                          self.server.wait, timeout=1)
         thread.kill()
 
     @mock.patch.object(server_module, 'LOG')
@@ -838,3 +888,22 @@ class TestServerLocking(test_utils.BaseTestCase):
 
         # We timed out. Ensure we didn't log anything.
         self.assertFalse(mock_log.warning.called)
+
+
+class TestRPCExposeDecorator(test_utils.BaseTestCase):
+
+    def foo(self):
+        pass
+
+    @rpc.expose
+    def bar(self):
+        """bar docstring"""
+        pass
+
+    def test_undecorated(self):
+        self.assertRaises(AttributeError, lambda: self.foo.exposed)
+
+    def test_decorated(self):
+        self.assertEqual(True, self.bar.exposed)
+        self.assertEqual("""bar docstring""", self.bar.__doc__)
+        self.assertEqual('bar', self.bar.__name__)
